/* dkill - Kill daemons started by dstart
   Copyright (C) 2012, 2014, 2019 William Haddon

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License,
   as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>

#ifndef RUNDIR
#define RUNDIR "/run"
#endif

#define SOCKET_SUBDIR "dstart"

#define EMSG_USAGE "USAGE: %s [-r] [-SIGNAL] NAME\n"
#define EMSG_SOCKET "%s: cannot create a UNIX stream socket: %s\n"
#define EMSG_CONNECT "%s: cannot connect to socket \"%s\": %s\n"
#define EMSG_NODAEMON "%s: no daemon \"%s\" managed by dstart\n"
#define EMSG_SEND "%s: cannot send data across UNIX socket \"%s\": %s\n"
#define EMSG_RECV "%s: cannot receive data from UNIX socket \"%s\": %s\n"

#define UNIX_PATH_MAX 108

void error(char *msg, ...) {
	va_list ap;
	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	exit(1);
}

int main(int argc, char **argv) {
	unsigned char action;
	int s, argpos;
	struct sockaddr_un addr;
	char *name, *p;
	/* Default action */
	action = 0x80 | SIGTERM;
	/* Process command line */
	for (argpos = 1; argpos < argc; argpos++) {
		if (*(argv[argpos]) != '-')
			break;
		if (isdigit(argv[argpos][1])) {
			action = (action & 0x80) | atoi(argv[argpos] + 1);
		} else {
			for (p = argv[argpos] + 1; *p; p++) {
				switch (*p) {
				case 'r':
					/* Let dstart restart the daemon */
					action &= 0x7f;
					break;
				default:
					error(EMSG_USAGE, argv[0]);
					break;
				}
			}
		}
	}
	if (argpos == argc)
		error(EMSG_USAGE, argv[0]);
	name = argv[argpos];
	/* Get a socket */
	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s < 0)
		error(EMSG_SOCKET, argv[0], strerror(errno));
	/* Connect the socket */
	addr.sun_family = AF_UNIX;
	snprintf(addr.sun_path, UNIX_PATH_MAX, RUNDIR"/"SOCKET_SUBDIR"/%s", name);
	if (connect(s, (struct sockaddr *)&addr, sizeof(struct sockaddr_un)) < 0) {
		if (errno == ENOENT)
			error(EMSG_NODAEMON, argv[0], name);
		error(EMSG_CONNECT, argv[0], addr.sun_path, strerror(errno));
	}
	/* Send the data */
	if (send(s, &action, 1, 0) < 0)
		error(EMSG_SEND, argv[0], addr.sun_path, strerror(errno));
	/* Wait for the response */
	if (recv(s, &action, 1, 0) < 0)
		error(EMSG_RECV, argv[0], addr.sun_path, strerror(errno));
	return 0;
}
