/* hwd - React to the addition and removal of hardware
   Copyright (C) 2019 William Haddon

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License,
   as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <poll.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/signalfd.h>
#include <sys/socket.h>
#include <sys/swap.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/mount.h>
#include <mntent.h>
#include <linux/netlink.h>
#include <arpa/inet.h>

#ifndef ETCDIR
#define ETCDIR "/etc"
#endif

#ifndef MOUNTSCRIPT
#define MOUNTSCRIPT ETCDIR"/init/mount"
#endif

#ifndef FSTAB
#define FSTAB ETCDIR"/fstab"
#endif

extern char **environ;

char *xstrdup(char *s, char *error) {
	char *result;
	result = strdup(s);
	if (!result) {
		fprintf(stderr, "%s: Out of memory\n", error);
		exit(1);
	}
	return result;
}

void *xmalloc(size_t size, char *error) {
	void *result;
	result = malloc(size);
	if (!result) {
		fprintf(stderr, "%s: Out of memory\n", error);
		exit(1);
	}
	return result;
}

void *xrealloc(void *ptr, size_t size, char *error) {
	void *result;
	result = realloc(ptr, size);
	if (!result) {
		fprintf(stderr, "%s: Out of memory\n", error);
		exit(1);
	}
	return result;
}

void xmount(char *fs, char *mountpoint, char *type, unsigned long flags,
		void *data, char *error) {
	if (mount(fs, mountpoint, type, flags, data) && errno != EBUSY) {
		fprintf(stderr, "%s: %s\n", error, strerror(errno));
		exit(1);
	}
}

pid_t vfe(char *proc, char **argv) {
	pid_t r;
	sigset_t mask, old;
	/* signals must be blocked for vfork to be safe */
	sigfillset(&mask);
	sigprocmask(SIG_BLOCK, &mask, &old);
	errno = 0;
	r = vfork();
	if (r) {
		if (errno)
			r = -1;
	} else {
		/* vfork returns the PID, but execve sets errno */
		if (execve(proc, argv, environ))
			_exit(1);
	}
	sigprocmask(SIG_SETMASK, &old, NULL);
	return r;
}

int ensure_directory(char *name) {
	struct stat st;
	if (!stat(name, &st)) {
		if (S_ISDIR(st.st_mode)) {
			return 0;
		} else {
			errno = ENOTDIR;
			return -1;
		}
	} else {
		if (errno != ENOENT)
			return -1;
		else
			return mkdir(name, 0755);
	}
}

struct scan_entry {
	char *name;
	DIR *dir;
};

int trigger_uevent_scan(char *dir, struct pollfd *fds,
		struct scan_entry *se) {
	int i;
	for (i = 2; i < 512; i++) {
		if (fds[i].fd < 0) {
			se[i].name = strdup(dir);
			if (!se[i].name) {
				printf("scanning %s: Out of memory\n", dir);
				return -1;
			}
			fds[i].fd = open(dir, O_RDONLY|O_NONBLOCK);
			if (fds[i].fd < 0) {
				free(se[i].name);
				se[i].name = NULL;
				printf("scanning %s: opening %s: %s\n", dir, dir,
					strerror(errno));
				return -1;
			}
			fds[i].events = POLLIN;
			fds[i].revents = 0;
			se[i].dir = NULL;
			return 0;
		}
	}
	errno = ENOBUFS;
	return -1;
}

void cleanup_uevent_scan(int idx, struct pollfd *fds, struct scan_entry *se) {
	closedir(se[idx].dir);
	free(se[idx].name);
	se[idx].dir = NULL;
	se[idx].name = NULL;
	fds[idx].fd = -1;
	fds[idx].events = 0;
	fds[idx].revents = 0;
}

int process_uevent_scan(int idx, struct pollfd *fds, struct scan_entry *se) {
	struct dirent *entry;
	FILE *f;
	char *child;
	size_t len;
	int r;
	len = strlen(se[idx].name);
	while (1) {
		if (se[idx].dir) {
			errno = 0;
			entry = readdir(se[idx].dir);
			if (!entry) {
				if (!errno) {
					cleanup_uevent_scan(idx, fds, se);
					return 0;
				} else if (errno == EAGAIN) {
					goto pause;
				} else {
					printf("scanning %s: %s\n", se[idx].name, strerror(errno));
					goto error;
				}
			} else if (entry->d_type == DT_DIR && strcmp(entry->d_name, ".") &&
					strcmp(entry->d_name, "..")) {
				child = malloc(strlen(se[idx].name) +
					strlen(entry->d_name) + 2);
				if (!child) {
					printf("scanning %s/%s: Out of memory\n", se[idx].name,
						entry->d_name);
					continue;
				}
				sprintf(child, "%s/%s", se[idx].name, entry->d_name);
				r = trigger_uevent_scan(child, fds, se);
				free(child);
				if (r) {
					if (errno == ENOBUFS) {
						seekdir(se[idx].dir, telldir(se[idx].dir) - 1);
						goto pause;
					} else {
						continue;
					}
				}
			} else if (!strcmp(entry->d_name, "uevent")) {
				child = malloc(strlen(se[idx].name) + 8);
				if (!child) {
					printf("scanning %s: Out of memory\n", se[idx].name);
					continue;
				}
				sprintf(child, "%s/uevent", se[idx].name);
				f = fopen(child, "w");
				free(child);
				if (!f) {
					printf("scanning %s: opening %s/uevent: %s\n",
						se[idx].name, se[idx].name, strerror(errno));
					continue;
				}
				if (fprintf(f, "add") < 0)
					printf("scanning %s: writing to %s/uevent: %s\n",
						se[idx].name, se[idx].name, strerror(errno));
				fclose(f);
				goto pause;
			}
		} else {
			se[idx].dir = fdopendir(fds[idx].fd);
			if (!se[idx].dir) {
				if (errno == EAGAIN) {
					goto pause;
				} else {
					printf("scanning %s: opening %s as directory: %s\n",
						se[idx].name, se[idx].name, strerror(errno));
					goto error;
				}
			}
		}
	}
pause:
	fds[idx].revents = 0;
	return 0;
error:
	cleanup_uevent_scan(idx, fds, se);
	return -1;
}

struct fslabel {
	char *source;
	char *uuid;
	char *label;
};

struct fs {
	struct fslabel label;
	char *target;
	char *fstype;
	int type;
	char *mountopts;
	int passno;
	int phase;
};

struct fs *fstable;
size_t fsts, fstl, curr_fs;
int curr_pass;
pid_t fs_script;

struct fslabel *swaptable;
size_t ssts, sstl, swapcount;

int ttys, late;

int compare_fs(struct fs *a, struct fs *b) {
	if (a->passno > b->passno)
		return 1;
	else if (a->passno == b->passno)
		return 0;
	else
		return -1;
}

void fsck_fs(int i) {
	char *argv[6];
	argv[0] = MOUNTSCRIPT;
	argv[1] = fstable[i].label.source;
	argv[2] = fstable[i].target;
	argv[3] = fstable[i].fstype;
	argv[4] = fstable[i].mountopts;
	argv[5] = NULL;
	fs_script = vfe(MOUNTSCRIPT, argv);
	if (fs_script < 0) {
		printf("executing "MOUNTSCRIPT": %s\n", strerror(errno));
	} else {
		fstable[i].phase = 2;
		curr_fs = i;
	}
}

void format_uuid(char *bin, char *s) {
	sprintf(s, "%.8x-%.4x-%.4x-%.4x-%.4x%.8x", ntohl(*(uint32_t *)bin),
		ntohs(*(uint16_t *)(bin + 4)), ntohs(*(uint16_t *)(bin + 6)),
		ntohs(*(uint16_t *)(bin + 8)), ntohs(*(uint16_t *)(bin + 10)),
		ntohl(*(uint32_t *)(bin + 12)));
}

int label_matches(char *filename, struct fslabel *label) {
	FILE *f;
	long offset;
	size_t i;
	uint16_t magic;
	uint32_t version;
	char *s;
	char buf[32];
	char buf2[32];
	char uuid[37];
	if (label->source && strcmp(filename, label->source))
		return 1;
	if (!label->uuid && !label->label)
		return 0;
	f = fopen(filename, "r");
	if (!f) {
		printf("opening %s to check identity: %s\n", filename,
			strerror(errno));
		return -1;
	}
	if (fseek(f, 1080, SEEK_SET))
		goto cleanup;
	if (!fread(&magic, 2, 1, f))
		goto error;
	if (ntohs(magic) == 0x53ef) { /* ext2/3/4 */
		if (fseek(f, 1100, SEEK_SET))
			goto cleanup;
		if (!fread(&version, 4, 1, f))
			goto error;
		if (ntohl(version) < 1)
			goto cleanup;
		if (label->uuid) {
			if (fseek(f, 1128, SEEK_SET))
				goto cleanup;
			if (!fread(buf, 16, 1, f))
				goto error;
			format_uuid(buf, uuid);
			if (strcasecmp(label->uuid, uuid))
				goto cleanup;
		}
		if (label->label) {
			if (fseek(f, 1144, SEEK_SET))
				goto cleanup;
			if (!fread(buf, 16, 1, f))
				goto error;
			if (strncmp(label->label, buf, 16))
				goto cleanup;
		}
	} else {
		if (fseek(f, 4086, SEEK_SET))
			goto cleanup;
		if (!fread(buf, 10, 1, f))
			goto error;
		if (!strncmp(buf, "SWAPSPACE2", 10)) { /* linux swap */
			if (label->uuid) {
				if (fseek(f, 1036, SEEK_SET))
					goto cleanup;
				if (!fread(buf, 16, 1, f))
					goto error;
				format_uuid(buf, uuid);
				if (strcasecmp(label->uuid, uuid))
					goto cleanup;
			}
			if (label->label)
				goto cleanup;
		} else {
			if (label->uuid)
				goto cleanup;
			if (label->label) {
				offset = 32768;
				s = label->label;
				for (i = 0; i < 32; i++) {
					if (*s) {
						buf2[i] = *s;
						s++;
					} else {
						buf2[i] = ' ';
					}
				}
				while (1) { /* ISO 9660 CD */
					if (fseek(f, offset, SEEK_SET))
						goto cleanup;
					if (!fread(buf, 6, 1, f))
						goto error;
					if (strncmp(buf + 1, "CD001", 5) || *buf == 255)
						goto cleanup;
					if (*buf == 1) {
						if (fseek(f, offset + 40, SEEK_SET))
							goto cleanup;
						if (!fread(buf, 32, 1, f))
							goto error;
						if (strncmp(buf2, buf, 32))
							goto cleanup;
						else
							goto success;
					}
					offset += 2048;
				}
			}
		}
	}
success:
	fclose(f);
	return 0;
error:
	printf("reading %s to check identity: %s\n", filename, strerror(errno));
	return -1;
cleanup:
	fclose(f);
	return 1;
}

int setup_dev(char *filename, mode_t mode, dev_t dev) {
	char *argv[3];
	size_t i;
	int maj, min;
	maj = major(dev);
	min = minor(dev);
	if (mknod(filename, mode, dev)) {
		printf("creating %s: %s\n", filename, strerror(errno));
		return -1;
	}
	if (S_ISBLK(mode)) {
		for (i = 0; i < fstl; i++) {
			if (!label_matches(filename, &(fstable[i].label))) {
				fstable[i].label.source = strdup(filename);
				if (!fstable[i].label.source) {
					printf("mounting %s: out of memory\n", filename);
					return -1;
				}
				if (curr_fs == fstl &&
						(late || fstable[i].passno <= curr_pass))
					fsck_fs(i);
				else
					fstable[i].phase = 1;
				return 0;
			}
		}
		for (i = 0; i < sstl; i++) {
			if (!label_matches(filename, swaptable + i)) {
				if (swapon(filename, 0)) {
					printf("enabling swap on %s: %s\n", filename,
						strerror(errno));
					return -1;
				}
				swapcount++;
				return 0;
			}
		}
	} else if (S_ISCHR(mode) && maj == 4 && min >= 1 && min <= 12) {
		ttys |= 1 << (min - 1);
	}
	return 0;
}

void create_dev(char *devname, mode_t mode, int major, int minor) {
	struct stat st;
	dev_t dev;
	char *filename, *s;
	int result;
	filename = malloc(strlen(devname) + 6);
	if (!filename) {
		printf("creating /dev/%s: Out of memory\n", devname);
		return;
	}
	sprintf(filename, "/dev/%s", devname);
	for (s = filename + 5; *s; s++) {
		if (*s == '/') {
			*s = 0;
			if (ensure_directory(filename)) {
				printf("creating %s: %s\n", filename, strerror(errno));
				free(filename);
				return;
			}
			*s = '/';
		}
	}
	dev = makedev(major, minor);
	if (stat(filename, &st)) {
		if (errno == ENOENT) {
			result = setup_dev(filename, mode, dev);
		} else {
			printf("creating %s: %s\n", filename, strerror(errno));
			result = -1;
		}
	} else if (st.st_mode != mode || st.st_rdev != dev) {
		printf("creating %s: file already exists\n", filename);
	} else {
		result = 0;
	}
	free(filename);
}

void remove_dev(char *devname) {
	struct stat st;
	char *filename;
	int result;
	filename = malloc(strlen(devname) + 6);
	if (!filename) {
		printf("deleting /dev/%s: Out of memory\n", devname);
		return;
	}
	sprintf(filename, "/dev/%s", devname);
	if (stat(filename, &st)) {
		if (errno != ENOENT)
			printf("deleting %s: %s\n", filename, strerror(errno));
	} else if (!S_ISBLK(st.st_mode) && !S_ISCHR(st.st_mode)) {
		printf("deleting %s: not a device file\n", filename);
	} else if (unlink(filename)) {
		printf("deleting %s: %s\n", filename, strerror(errno));
	}
	free(filename);
}

int process_kvp(char **buf, ssize_t *len, char **key, char **value) {
	char *s;
	int result;
	*key = *buf;
	result = -1;
	for (; *len > 0; (*buf)++, (*len)--) {
		switch (**buf) {
		case '=':
			**buf = 0;
			result = 0;
			*value = *buf + 1;
			break;
		case 0:
			(*buf)++;
			(*len)--;
			return result;
		}
	}
	return -1;
}

int process_ignored(char *buf, ssize_t len) {
	char *key, *value;
	int seqnum;
	/*printf("unrecognized netlink packet (bind or unbind)\n");*/
	seqnum = -1;
	while (len > 0)
		if (!process_kvp(&buf, &len, &key, &value) && !strcmp(key, "SEQNUM"))
			seqnum = atoi(value);
	return seqnum;
}

#define END 0
#define SUBSYS 1
#define NAME 2

struct {
	int type;
	char *str;
	mode_t mode;
} rules[] = {
	{SUBSYS, "tty", 0600 | S_IFCHR},
	{SUBSYS, "block", 0600 | S_IFBLK},
	{SUBSYS, "input", 0600 | S_IFCHR},
	{SUBSYS, "vc", 0600 | S_IFCHR},
	{SUBSYS, "drm", 0600 | S_IFCHR},
	{SUBSYS, "rtc", 0600 | S_IFCHR},
	{SUBSYS, "sound", 0666 | S_IFCHR},
	{SUBSYS, "scsi_generic", 0600 | S_IFCHR},
	{NAME, "zero", 0666 | S_IFCHR},
	{NAME, "full", 0666 | S_IFCHR},
	{NAME, "urandom", 0666 | S_IFCHR},
	{NAME, "random", 0666 | S_IFCHR},
	{NAME, "null", 0666 | S_IFCHR},
	{SUBSYS, "mem", 0600 | S_IFCHR},
	{NAME, "vga_arbiter", 0600 | S_IFCHR},
	{NAME, "loop-control", 0600 | S_IFCHR},
	{END, NULL, 0}
};

mode_t find_mode(char *devname, char *subsys) {
	int i;
	for (i = 0; rules[i].type != END; i++) {
		switch (rules[i].type) {
		case NAME:
			if (!strcmp(devname, rules[i].str))
				return rules[i].mode;
			break;
		case SUBSYS:
			if (!strcmp(subsys, rules[i].str))
				return rules[i].mode;
			break;
		}
	}
	printf("can not find rule for device %s subsystem %s\n", devname, subsys);
	return 0;
}

int process_add(char *buf, ssize_t len) {
	char *key, *value, *devname, *subsys;
	int major, minor, seqnum, mask, i;
	mode_t mode;
	mask = 0;
	seqnum = -1;
	while (len > 0) {
		if (!process_kvp(&buf, &len, &key, &value)) {
			if (!strcmp(key, "MAJOR")) {
				major = atoi(value);
				mask |= 1;
			} else if (!strcmp(key, "MINOR")) {
				minor = atoi(value);
				mask |= 2;
			} else if (!strcmp(key, "SEQNUM")) {
				seqnum = atoi(value);
				mask |= 4;
			} else if (!strcmp(key, "DEVNAME")) {
				devname = value;
				mask |= 8;
			} else if (!strcmp(key, "SUBSYSTEM")) {
				subsys = value;
				mask |= 16;
			}
		}
	}
	if (mask == 31) {
		mode = find_mode(devname, subsys);
		if (mode)
			create_dev(devname, mode, major, minor);
	} else {
		/*printf("unrecognized netlink packet (add)\n");*/
	}
	return seqnum;
}

int process_remove(char *buf, ssize_t len) {
	char *key, *value, *devname;
	int seqnum, mask;
	seqnum = -1;
	devname = NULL;
	mask = 0;
	while (len > 0) {
		if (!process_kvp(&buf, &len, &key, &value)) {
			if (!strcmp(key, "DEVNAME")) {
				devname = value;
				mask |= 1;
			} else if (!strcmp(key, "SEQNUM")) {
				seqnum = atoi(value);
				mask |= 2;
			}
		}
	}
	if (mask == 3) {
		remove_dev(devname);
	} else {
		/*printf("unrecognized netlink packet (remove)\n");*/
	}
	return seqnum;
}

int process_packet(char *buf, ssize_t len) {
	if (len >= 4 && !strncmp(buf, "add@", 4)) {
		return process_add(buf + 4, len - 4);
	} else if (len >= 7 && !strncmp(buf, "remove@", 7)) {
		return process_remove(buf + 7, len - 7);
	} else if ((len >= 5 && !strncmp(buf, "bind@", 5))) {
		return process_ignored(buf + 5, len - 5);
	} else if ((len >= 7 && !strncmp(buf, "unbind@", 7)) ||
			(len >= 7 && !strncmp(buf, "change@", 7))) {
		return process_ignored(buf + 7, len - 7);
	} else {
		/*printf("unrecognized netlink packet\n");*/
		return -1;
	}
}

void initialize_netlink(struct pollfd *fds, struct scan_entry *se) {
	struct sockaddr_nl sa;
	int s;
	s = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT);
	if (s < 0) {
		fprintf(stderr, "initializing netlink socket: %s\n", strerror(errno));
		exit(1);
	}
	fds[0].fd = s;
	fds[0].events = POLLIN;
	fds[0].revents = 0;
	se[0].name = NULL;
	se[0].dir = NULL;
	sa.nl_family = AF_NETLINK;
	sa.nl_pad = 0;
	sa.nl_pid = 0;
	sa.nl_groups = 1;
	if (bind(s, (struct sockaddr *)(&sa), sizeof(sa))) {
		fprintf(stderr, "connecting to netlink socket: %s\n",
			strerror(errno));
		exit(1);
	}
}

char pktbuf[1024];

int process_netlink(struct pollfd *fds, struct scan_entry *se, int seqnum) {
	struct sockaddr_nl sa;
	socklen_t len;
	ssize_t r;
	int next;
	errno = 0;
	len = sizeof(struct sockaddr_nl);
	r = recvfrom(fds[0].fd, pktbuf, 1024, 0, (struct sockaddr *)(&sa), &len);
	if (r < 0) {
		printf("receiving from netlink socket: %s\n", strerror(errno));
	} else if (sa.nl_pid || sa.nl_groups != 1) {
		printf("receiving from netlink socket: invalid origin\n");
	} else {
		next = process_packet(pktbuf, r);
		if (next >= 0) {
			if (seqnum >= 0 && seqnum + 1 < next)
				trigger_uevent_scan("/sys", fds, se);
			seqnum = next;
		}
	}
	return seqnum;
}

void parse_label(struct fslabel *label, char *s) {
	label->source = NULL;
	label->uuid = NULL;
	label->label = NULL;
	if (!strncmp(s, "UUID=", 5)) {
		label->uuid = xstrdup(s + 5, "processing "FSTAB);
	} else if (!strncmp(s, "LABEL=", 6)) {
		label->label = xstrdup(s + 6, "processing "FSTAB);
	} else {
		label->source = xstrdup(s, "processing "FSTAB);
	}
}

int mounted_system_fss() {
	FILE *f;
	struct mntent *me;
	int result;
	result = 0;
	f = setmntent("/proc/mounts", "r");
	if (!f) {
		fprintf(stderr, "opening /proc/mounts: %s\n", strerror(errno));
		exit(1);
	}
	errno = 0;
	while (me = getmntent(f)) {
		if (!strcmp(me->mnt_dir, "/dev"))
			result |= 1;
		else if (!strcmp(me->mnt_dir, "/sys"))
			result |= 2;
		else if (!strcmp(me->mnt_dir, "/dev/pts"))
			result |= 4;
	}
	if (errno) {
		fprintf(stderr, "reading /proc/mounts: %s\n", strerror(errno));
		exit(1);
	}
	endmntent(f);
	return result;
}

void read_fstab() {
	FILE *f;
	struct mntent *me;
	fsts = 16;
	fstl = 0;
	ssts = 16;
	sstl = 0;
	fstable = xmalloc(fsts * sizeof(struct fs), "processing "FSTAB);
	swaptable = xmalloc(ssts * sizeof(struct fslabel), "processing "FSTAB);
	f = setmntent(FSTAB, "r");
	if (!f) {
		fprintf(stderr, "opening "FSTAB": %s\n", strerror(errno));
		exit(1);
	}
	errno = 0;
	while (me = getmntent(f)) {
		if (!hasmntopt(me, "noauto")) {
			if (hasmntopt(me, "sw")) {
				if (sstl == ssts) {
					ssts *= 2;
					swaptable = xrealloc(swaptable,
						ssts * sizeof(struct fslabel), "processing "FSTAB);
				}
				parse_label(swaptable + sstl, me->mnt_fsname);
				sstl++;
			} else {
				if (fstl == fsts) {
					fsts *= 2;
					fstable = xrealloc(fstable, fsts * sizeof(struct fs),
						"processing "FSTAB);
				}
				parse_label(&(fstable[fstl].label), me->mnt_fsname);
				fstable[fstl].target = xstrdup(me->mnt_dir,
					"processing "FSTAB);
				fstable[fstl].fstype = xstrdup(me->mnt_type,
					"processing "FSTAB);
				fstable[fstl].mountopts = xstrdup(me->mnt_opts,
					"processing "FSTAB);
				if (me->mnt_passno >= 0)
					fstable[fstl].passno = me->mnt_passno;
				else
					fstable[fstl].passno = 0;
				fstable[fstl].phase = 0;
				fstl++;
			}
		}
	}
	if (errno) {
		fprintf(stderr, "reading "FSTAB": %s\n", strerror(errno));
		exit(1);
	}
	endmntent(f);
	qsort(fstable, fstl, sizeof(struct fs),
		((int(*)(const void*, const void*))compare_fs));
	curr_fs = fstl;
	fs_script = 0;
	if (fstl)
		curr_pass = fstable[0].passno;
	else
		curr_pass = -1;
	swapcount = 0;
}

/* 0 - waiting; 1 - nothing to do; 2 - something done */
int next_fsck() {
	size_t i;
	int next_pass;
	if (curr_pass < 0)
		return 1;
	next_pass = 1;
	for (i = 0; i < fstl && fstable[curr_fs].passno <= curr_pass; i++) {
		if (fstable[i].phase == 1 || (late && fstable[i].phase == 0)) {
			fsck_fs(i);
			return 2;
		} else if (fstable[i].phase == 0) {
			next_pass = 0;
		}
	}
	if (next_pass) {
		if (i < fstl)
			curr_pass = fstable[i].passno;
		else
			curr_pass = -1;
	}
	return next_pass;
}

void signal_done() {
	late = 1;
	if (kill(getppid(), SIGTERM)) {
		fprintf(stderr, "signaling initialization completion to dstart: %s\n",
			strerror(errno));
		exit(1);
	}
}

void sighandler(struct signalfd_siginfo *ss) {
	pid_t pid;
	int status;
	switch (ss->ssi_signo) {
	case SIGCHLD:
		pid = waitpid(-1, &status, WNOHANG);
		while (pid > 0) {
			if (pid == fs_script) {
				fstable[curr_fs].phase = 3;
				while (next_fsck() == 1 && curr_pass >= 0);
				fs_script = 0;
				curr_fs = fstl;
			}
			pid = waitpid(-1, &status, WNOHANG);
		}
		break;
	case SIGALRM:
		if (!late)
			signal_done();
		break;
	}
}

char *get_root_fs() {
	FILE *f;
	char *s, *result;
	size_t size, pos;
	int ch, mode;
	result = NULL;
	size = 256;
	f = fopen("/proc/cmdline", "r");
	if (!f) {
		fprintf(stderr, "opening /proc/cmdline: %s\n", strerror(errno));
		exit(1);
	}
	s = xmalloc(size, "reading /proc/cmdline");
	mode = 0;
	for (ch = fgetc(f); ch >= 0; ch = fgetc(f)) {
		if ((ch != '\n' || ch != '\t' || ch != '\r' || ch != ' ') &&
				mode == 0) {
			s[0] = ch;
			pos = 1;
			mode = 1;
		} else if ((ch == '\n' || ch == '\t' || ch == '\r' || ch == ' ' ||
				ch == '=') && mode == 1) {
			s[pos] = 0;
			pos = 0;
			if (ch == '=') {
				if (!strcmp(s, "root"))
					mode = 2;
				else
					mode = 3;
			} else {
				mode = 0;
			}
		} else if ((ch == '\n' || ch == '\t' || ch == '\r' || ch == ' ') &&
				(mode == 2 || mode == 3)) {
			s[pos] = 0;
			if (mode == 2)
				result = xstrdup(s, "reading /proc/cmdline");
			pos = 0;
			mode = 0;
		} else {
			s[pos] = ch;
			if (pos == size) {
				size *= 2;
				s = xrealloc(s, size, "reading /proc/cmdline");
			}
			pos++;
		}
	}
	if (mode == 2)
		result = xstrdup(s, "reading /proc/cmdline");
	fclose(f);
	return result;
}

int main() {
	struct pollfd *fds;
	struct scan_entry *se;
	struct signalfd_siginfo ss;
	struct stat st;
	char *rootfs;
	sigset_t mask;
	ssize_t r;
	int i, msf, seqnum, ct;
	late = 0;
	fds = xmalloc(512 * sizeof(struct pollfd), "initializing");
	se = xmalloc(512 * sizeof(struct scan_entry), "initializing");
	umask(0);
	ttys = 0;
	errno = 0;
	if (stat("/proc/mounts", &st)) {
		if (errno == ENOENT) {
			xmount(NULL, "/proc", "proc", MS_MGC_VAL, NULL, "mounting /proc");
		} else {
			fprintf(stderr, "checking for /proc: %s\n", strerror(errno));
			return 1;
		}
	}
	msf = mounted_system_fss();
	if (!(msf & 1))
		xmount(NULL, "/dev", "tmpfs", MS_MGC_VAL, "mode=755", "mounting /dev");
	if (!(msf & 2))
		xmount(NULL, "/sys", "sysfs", MS_MGC_VAL, NULL, "mounting /sys");
	if (!(msf & 4)) {
		if (ensure_directory("/dev/pts")) {
			fprintf(stderr, "creating /dev/pts: %s\n", strerror(errno));
			return 1;
		} else {
			xmount(NULL, "/dev/pts", "devpts", MS_MGC_VAL, NULL,
				"mounting /dev/pts");
		}
	}
	if (lstat("/dev/fd", &st)) {
		if (errno == ENOENT) {
			if (symlink("/proc/self/fd", "/dev/fd"))
				printf("creating /dev/fd: %s\n", strerror(errno));
		} else {
			printf("checking /dev/fd: %s\n", strerror(errno));
		}
	} else if (!S_ISLNK(st.st_mode)) {
		printf("creating /dev/fd: file already exists\n");
	}
	if (symlink("/proc/self/fd/0", "/dev/stdin"))
		printf("creating /dev/stdin: %s\n", strerror(errno));
	if (symlink("/proc/self/fd/1", "/dev/stdout"))
		printf("creating /dev/stdout: %s\n", strerror(errno));
	if (symlink("/proc/self/fd/2", "/dev/stderr"))
		printf("creating /dev/stderr: %s\n", strerror(errno));
	rootfs = get_root_fs();
	if (rootfs && symlink(rootfs, "/dev/root"))
		printf("creating /dev/root: %s\n", strerror(errno));
	read_fstab();
	initialize_netlink(fds, se);
	sigemptyset(&mask);
	sigaddset(&mask, SIGCHLD);
	sigaddset(&mask, SIGALRM);
	sigprocmask(SIG_BLOCK, &mask, NULL);
	alarm(10);
	fds[1].fd = signalfd(-1, &mask, SFD_NONBLOCK);
	if (!fds[1].fd) {
		fprintf(stderr, "establish signal handler: %s\n", strerror(errno));
		return 1;
	}
	fds[1].events = POLLIN;
	fds[1].revents = 0;
	se[1].name = NULL;
	se[1].dir = NULL;
	for (i = 2; i < 512; i++) {
		fds[i].fd = -1;
		fds[i].events = 0;
		fds[i].revents = 0;
		se[i].name = NULL;
		se[i].dir = NULL;
	}
	seqnum = -1;
	trigger_uevent_scan("/sys", fds, se);
	while (1) {
		if (!late && curr_pass < 0 && sstl == swapcount && ttys == 0xfff)
			signal_done();
		ct = poll(fds, 512, -1);
		if (fds[0].revents & POLLIN) {
			ct--;
			seqnum = process_netlink(fds, se, seqnum);
		} else if (fds[1].revents & POLLIN) {
			r = read(fds[1].fd, &ss, sizeof(struct signalfd_siginfo));
			while (r == sizeof(struct signalfd_siginfo)) {
				sighandler(&ss);
				r = read(fds[1].fd, &ss, sizeof(struct signalfd_siginfo));
			}
			if (r > 0)
				printf("incomplete signal packet\n");
			else if (r < 0 && errno != EAGAIN)
				printf("reading signal packet: %s\n", strerror(errno));
			fds[1].revents = 0;
		} else {
			for (i = 2; i < 512; i++) {
				if (fds[i].revents) {
					ct--;
					process_uevent_scan(i, fds, se);
				}
			}
		}
	}
	return 0;
}
